import asyncio
import base64
import io
import json
import os
import time
import uuid
from enum import Enum
from typing import Any, Optional, Tuple

import numpy as np
from pydub import AudioSegment

from spoken.base import SpeechToSpeechHarness


class NovaSpeechToSpeechHarness(SpeechToSpeechHarness):
    """
    Amazon Nova Sonic (https://aws.amazon.com/ai/generative-ai/nova/speech/)

    - amazon.nova-sonic-v1:0
    """

    class Model(Enum):
        AMAZON_NOVA_SONIC_V1_0 = "amazon.nova-sonic-v1:0"

    input_audio_sample_rate: Optional[int] = 16000
    output_audio_sample_rate: Optional[int] = 24000
    audio_token_frame_rate: Optional[int] = 25

    bedrock_client: Any
    stream: Optional[Any]
    is_active: bool
    role: Optional[str]

    def __init__(
        self,
        model: Model,
        source_audio_signal: np.ndarray,
        system_prompt: Optional[str] = None,
        temperature: float = 0.7,
    ):
        super().__init__(model, source_audio_signal, system_prompt, temperature)

        from aws_sdk_bedrock_runtime.client import BedrockRuntimeClient
        from aws_sdk_bedrock_runtime.config import (Config,
                                                    HTTPAuthSchemeResolver,
                                                    SigV4AuthScheme)
        from smithy_aws_core.credentials_resolvers.environment import \
            EnvironmentCredentialsResolver

        region = os.environ.get("AWS_REGION", "us-east-1")
        config = Config(
            endpoint_uri=f"https://bedrock-runtime.{region}.amazonaws.com",
            region=region,
            aws_credentials_identity_resolver=EnvironmentCredentialsResolver(),
            http_auth_scheme_resolver=HTTPAuthSchemeResolver(),
            http_auth_schemes={"aws.auth#sigv4": SigV4AuthScheme()},
        )
        self.bedrock_client = BedrockRuntimeClient(config=config)
        self.stream = None
        self.is_active = False

        self.role = None

    async def run(self) -> Tuple[Optional[str], str, AudioSegment]:
        from aws_sdk_bedrock_runtime.client import \
            InvokeModelWithBidirectionalStreamOperationInput

        self.session_id = str(uuid.uuid4())
        self.prompt_name = str(uuid.uuid4())
        self.text_content_id = str(uuid.uuid4())
        self.audio_content_id = str(uuid.uuid4())

        self.is_active = True
        self.stream = await self.bedrock_client.invoke_model_with_bidirectional_stream(
            InvokeModelWithBidirectionalStreamOperationInput(model_id=self.model_name)
        )

        response_task = asyncio.create_task(self._process_messages())

        # 1. sessionStart
        await self._send_message({
            "event": {
                "sessionStart": {
                    "inferenceConfiguration": {
                        "maxTokens": 4096,
                        "topP": 0.9,
                        "temperature": self.temperature,
                    }
                }
            }
        })

        # 2. promptStart
        await self._send_message({
            "event": {
                "promptStart": {
                    "promptName": self.prompt_name, 
                    "textOutputConfiguration": {
                        "mediaType": "text/plain",
                        "enabled": True,
                    },
                    "audioOutputConfiguration": {
                        "audioType": "SPEECH",
                        "encoding": "base64",
                        "mediaType": "audio/lpcm",
                        "sampleRateHertz": self.output_audio_sample_rate,
                        "sampleSizeBits": 16,
                        "channelCount": 1,
                        "voiceId": "matthew",
                    },
                    "toolUseOutputConfiguration": {"mediaType": "application/json"},
                    "toolConfiguration": {"tools": []},
                }
            }
        })

        system_prompt = (
            self.system_prompt
            or "You are a friendly assistant. The user and you will engage in a spoken dialog "
            "exchanging the transcripts of a natural real-time conversation. Keep your responses short, "
            "generally two or three sentences for chatty scenarios."
        )

        # 3. contentStart for system prompt
        await self._send_message({
            "event": {
                "contentStart": {
                    "promptName": self.prompt_name,
                    "contentName": self.text_content_id,
                    "role": "SYSTEM",
                    "type": "TEXT",
                    "interactive": True,
                    "textInputConfiguration": {"mediaType": "text/plain"},
                }
            }
        })

        # 4. textInput for system prompt
        await self._send_message({
            "event": {
                "textInput": {
                    "promptName": self.prompt_name,
                    "contentName": self.text_content_id,
                    "content": system_prompt,
                }
            }
        })

        # 5. contentEnd for system prompt
        await self._send_message({
            "event": {
                "contentEnd": {
                    "promptName": self.prompt_name,
                    "contentName": self.text_content_id,
                }
            }
        })

        # 6. contentStart for input audio
        await self._send_message({
            "event": {
                "contentStart": {
                    "promptName": self.prompt_name,
                    "contentName": self.audio_content_id,
                    "type": "AUDIO",
                    "interactive": True,
                    "role": "USER",
                    "audioInputConfiguration": {
                        "mediaType": "audio/lpcm",
                        "sampleRateHertz": self.input_audio_sample_rate,
                        "sampleSizeBits": 16,
                        "encoding": "base64",
                        "channelCount": 1,
                        "audioType": "SPEECH",
                    },
                }
            }
        })

        # 7. audioInput for input audio
        await self._send_message({
            "event": {
                "audioInput": {
                    "promptName": self.prompt_name,
                    "contentName": self.audio_content_id,
                    "content": self.input_audio_base64,
                }
            }
        })

        # 8. simulate 200ms of silence for automatic turn detection to trigger END_TURN for USER
        quiet_chunk_base64 = base64.b64encode(b"\x00" * 1024).decode("utf-8")
        for _ in range(200):
            await self._send_message({
                "event": {
                    "audioInput": {
                        "promptName": self.prompt_name,
                        "contentName": self.audio_content_id,
                        "content": quiet_chunk_base64,
                    }
                }
            })
            await asyncio.sleep(0.01)

        self.input_audio_sent_time = time.perf_counter()
        self.logger.info("Input audio fully sent at: {}", self.input_audio_sent_time)

        self.logger.info("Waiting for server to complete output...")
        timeout_seconds = 10
        start_time = time.time()

        while not self.is_complete and (time.time() - start_time) < timeout_seconds:
            await asyncio.sleep(0.1)

        if not self.is_complete:
            self.logger.warning("Timeout waiting for Nova response after {} seconds", timeout_seconds)
            self.is_complete = True

        if self.is_active:
            # 9. contentEnd for input audio
            await self._send_message({
                "event": {
                    "contentEnd": {
                        "promptName": self.prompt_name,
                        "contentName": self.audio_content_id,
                    }
                }
            })

            # 10. promptEnd
            await self._send_message({
                "event": {
                    "promptEnd": {
                        "promptName": self.prompt_name,
                    }
                }
            })

            # 11. sessionEnd
            await self._send_message({"event": {"sessionEnd": {}}})
            await self.stream.input_stream.close()

        if response_task and not response_task.done():
            self.is_active = False
            try:
                await asyncio.wait_for(response_task, timeout=2.0)
            except asyncio.TimeoutError:
                self.logger.debug("Response task didn't stop naturally, cancelling...")
                response_task.cancel()
                try:
                    await response_task
                except asyncio.CancelledError:
                    pass

        self.is_active = False

        self.output_audio = AudioSegment.from_raw(
            io.BytesIO(self.output_audio_bytes),
            sample_width=2,
            frame_rate=self.output_audio_sample_rate,
            channels=1,
        )

        return self.input_transcription, self.output_transcription, self.output_audio

    async def _send_message(self, event_dict):
        from aws_sdk_bedrock_runtime.models import (
            BidirectionalInputPayloadPart,
            InvokeModelWithBidirectionalStreamInputChunk)

        event_json = json.dumps(event_dict)
        self.logger.debug("Sending event: {}", event_json)
        event = InvokeModelWithBidirectionalStreamInputChunk(
            value=BidirectionalInputPayloadPart(bytes_=event_json.encode("utf-8"))
        )
        await self.stream.input_stream.send(event)

    async def _process_messages(self):
        try:
            self.logger.info("Starting response processing")
            while self.is_active:
                try:
                    output = await self.stream.await_output()
                    result = await output[1].receive()

                    if result.value and result.value.bytes_:
                        response_data = result.value.bytes_.decode("utf-8")
                        try:
                            json_data = json.loads(response_data)
                            self._handle_event(json_data)
                        except json.JSONDecodeError:
                            self.logger.debug("Non-JSON response: {}", response_data)
                    else:
                        self.logger.debug("Empty response received")
                except StopAsyncIteration:
                    self.logger.info("Stream ended")
                    break
                except Exception as e:
                    if "No events to transform were found" not in str(e):
                        self.logger.error("Error receiving response: {}", e)
                    break
        except Exception as e:
            if self.is_active:
                self.logger.error("Error processing responses: {}", e)
                self.is_complete = True

    def _handle_event(self, data):
        self.logger.debug("Received event: {}", json.dumps(data, indent=2))

        if "event" in data:
            event = data["event"]

            if "contentStart" in event:
                self.logger.info("Content start received")
                self.is_ready = True
                self.role = event["contentStart"].get("role")

            elif "audioOutput" in event:
                audio_output = event["audioOutput"]
                content_id = audio_output.get("contentId", audio_output.get("contentName", "unknown"))
                self.logger.info("Audio output received for: {} (type: {})", content_id, audio_output.get("type", "unknown"))

                if "content" in audio_output:
                    if self.first_output_token_time is None:
                        self.first_output_token_time = time.perf_counter()
                        self.logger.info("First output token received at: {}", self.first_output_token_time)
                    
                    audio_content = audio_output["content"]
                    if isinstance(audio_content, str):
                        try:
                            audio_bytes = base64.b64decode(audio_content)
                            self.output_audio_bytes += audio_bytes
                            self.logger.debug("Received audio chunk: {} bytes", len(audio_bytes))
                        except Exception as e:
                            self.logger.error("Failed to decode audio content: {}", e)
                    else:
                        self.logger.warning("Unexpected audio content format: {}", type(audio_content))

            elif "textOutput" in event:
                text_output = event["textOutput"]
                content_id = text_output.get("contentId", text_output.get("contentName", "unknown"))
                self.logger.info("Text output received for: {} (type: {})", content_id, text_output.get("type", "unknown"))

                if "content" in text_output:
                    text_content = text_output["content"]
                    if isinstance(text_content, str):
                        if self.role == "USER":
                            self.input_transcription += text_content
                            self.logger.info("Input transcription: {}", self.input_transcription)
                        else:
                            self.output_transcription += text_content
                            self.logger.info("Output transcription: {}", self.output_transcription)

            elif "contentEnd" in event:
                content_end = event["contentEnd"]
                content_id = content_end.get("contentId", content_end.get("contentName", "unknown"))
                self.logger.info("Content end received for: {} (type: {})", content_id, content_end.get("type", "unknown"))
                if content_end.get("type") == "AUDIO" and content_end.get("stopReason") == "END_TURN":
                    self.logger.info("Audio content end received - output complete")
                    self.is_complete = True

            elif "promptEnd" in event:
                self.logger.info("Prompt end received")

            elif "sessionEnd" in event:
                self.logger.info("Session end received")
                self.is_complete = True

            elif "usageEvent" in event:
                usage_event = event["usageEvent"]
                if "details" in usage_event and "total" in usage_event["details"]:
                    total = usage_event["details"]["total"]
                    if "input" in total and "speechTokens" in total["input"]:
                        self.input_audio_tokens = total["input"]["speechTokens"]
                        self.logger.info("Input audio tokens: {}", self.input_audio_tokens)
                    if "output" in total and "speechTokens" in total["output"]:
                        self.output_audio_tokens = total["output"]["speechTokens"]
                        self.logger.info("Output audio tokens: {}", self.output_audio_tokens)

            elif "error" in event:
                self.logger.error("Error received: {}", json.dumps(event["error"], indent=2))
                self.is_complete = True

        elif "error" in data:
            self.logger.error("Server error: {}", json.dumps(data, indent=2))
            self.is_complete = True


if __name__ == "__main__":
    from pathlib import Path

    # os.environ["AWS_ACCESS_KEY_ID"] = "your-aws-access-key-id"
    # os.environ["AWS_SECRET_ACCESS_KEY"] = "your-aws-secret-access-key"
    # os.environ["AWS_SESSION_TOKEN"] = "your-aws-session-token"

    harness = NovaSpeechToSpeechHarness.from_file(
        NovaSpeechToSpeechHarness.Model.AMAZON_NOVA_SONIC_V1_0,
        Path("./examples/scooby.wav")
    )

    input_transcription, output_transcription, output_audio = asyncio.run(harness.run())
    print(f"Input transcription: {input_transcription}")
    print(f"Output transcription: {output_transcription}")
    output_audio.export("./nova.wav", format="wav")



import asyncio
import base64
import io
import json
import os
from enum import Enum
from typing import Optional, Tuple
import time

import numpy as np
import websockets
from pydub import AudioSegment

from spoken.base import SpeechToSpeechHarness


class OpenAISpeechToSpeechHarness(SpeechToSpeechHarness):
    """
    OpenAI Realtime Audio (https://platform.openai.com/docs/guides/realtime)
    - gpt-4o-realtime-preview-2024-12-17
    """

    class Model(Enum):
        GPT_4O_REALTIME_PREVIEW_2024_12_17 = "gpt-4o-realtime-preview-2024-12-17"

    input_audio_sample_rate: Optional[int] = 24000
    output_audio_sample_rate: Optional[int] = 24000
    audio_token_frame_rate: Optional[int] = 10  # 100ms per token

    def __init__(
        self,
        model: Model,
        source_audio_signal: np.ndarray,
        system_prompt: Optional[str] = None,
        temperature: float = 0.8,
    ):
        super().__init__(model, source_audio_signal, system_prompt, temperature)

        api_key = os.environ.get("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("OPENAI_API_KEY environment variable is required")

        self.realtime_base_url = f"wss://api.openai.com/v1/realtime?model={self.model_name}"
        self.realtime_headers = [
            ("Authorization", f"Bearer {api_key}"),
            ("OpenAI-Beta", "realtime=v1"),
        ]

    async def run(self) -> Tuple[Optional[str], str, AudioSegment]:
        async with websockets.connect(self.realtime_base_url, additional_headers=self.realtime_headers) as websocket:
            setup_message = {
                "type": "session.update",
                "session": {
                    "input_audio_format": "pcm16",
                    "output_audio_format": "pcm16",
                    "temperature": self.temperature,
                    "turn_detection": None,
                    "speed": 1.5,
                }
            }
            await websocket.send(json.dumps(setup_message))
            self.logger.debug("Sending setup message: {}", json.dumps(setup_message, indent=2))

            while not self.is_ready:
                try:
                    message = await asyncio.wait_for(websocket.recv(), timeout=1.0)
                    self._process_message(message)
                except asyncio.TimeoutError:
                    continue

            self.logger.debug("Sending audio input")
            await websocket.send(json.dumps({
                "type": "conversation.item.create",
                "item": {
                    "type": "message",
                    "role": "user",
                    "content": [
                        {"type": "input_audio", "audio": self.input_audio_base64}
                    ],
                },
            }))

            kwargs = {}
            if self.system_prompt:
                kwargs["instructions"] = self.system_prompt
            await websocket.send(json.dumps({"type": "response.create", "response": kwargs}))
            self.input_audio_sent_time = time.perf_counter()
            self.logger.info("Input audio sent at: {}", self.input_audio_sent_time)

            while not self.is_complete:
                try:
                    message = await asyncio.wait_for(websocket.recv(), timeout=1.0)
                    self._process_message(message)
                except asyncio.TimeoutError:
                    continue

        self.output_audio = AudioSegment.from_raw(
            io.BytesIO(self.output_audio_bytes),
            sample_width=2,
            frame_rate=self.output_audio_sample_rate,
            channels=1,
        )

        # OpenAI doesn't do input transcription
        return None, self.output_transcription, self.output_audio

    def _process_message(self, message: str):
        data = json.loads(message)
        self.logger.debug("Received event: {}", json.dumps(data, indent=2))
        if data.get("type") == "session.created":
            self.logger.info("Session created")
            self.is_ready = True
        elif data.get("type") == "error":
            self.logger.error("Error: {}", json.dumps(data, indent=2))
        elif data.get("type") == "response.done":
            self.output_transcription = data["response"]["output"][0]["content"][0]["transcript"]
            self.is_complete = True
            self.logger.info("Transcription: {}", self.output_transcription)
            self.input_audio_tokens = data["response"]["usage"]["input_token_details"]["audio_tokens"]
            self.output_audio_tokens = data["response"]["usage"]["output_token_details"]["audio_tokens"]
            self.logger.info("Input tokens: {}", self.input_audio_tokens)
            self.logger.info("Output tokens: {}", self.output_audio_tokens)
        elif data.get("type") == "response.audio.delta":
            if self.first_output_token_time is None:
                self.first_output_token_time = time.perf_counter()
                self.logger.info("First output token received at: {}", self.first_output_token_time)

            delta_b64 = data["delta"]
            delta_bytes = base64.b64decode(delta_b64)
            self.output_audio_bytes += delta_bytes

if __name__ == "__main__":
    from pathlib import Path

    # os.environ["OPENAI_API_KEY"] = "your-openai-api-key"

    harness = OpenAISpeechToSpeechHarness.from_file(
        OpenAISpeechToSpeechHarness.Model.GPT_4O_REALTIME_PREVIEW_2024_12_17,
        Path("./examples/scooby.wav")
    )

    input_transcription, output_transcription, output_audio = asyncio.run(harness.run())
    print(f"Input transcription: {input_transcription}")
    print(f"Output transcription: {output_transcription}")
    output_audio.export("./openai.wav", format="wav")



import inspect
import sys
from pathlib import Path
from typing import Optional, Union

from pydub import AudioSegment

from .base import SpeechToSpeechHarness, SpeechToSpeechHarnessMeta
from .models.gemini import GeminiSpeechToSpeechHarness
from .models.nova import NovaSpeechToSpeechHarness
from .models.openai import OpenAISpeechToSpeechHarness

def spoken(
        model_name: str,
        input_audio: Union[str, Path, AudioSegment],
        system_prompt: Optional[str] = None,
        temperature: Optional[float] = None, # TODO: use the default for the model
    ) -> SpeechToSpeechHarness:
    cls = SpeechToSpeechHarnessMeta.name_to_harness(model_name)
    model = cls.Model(model_name)

    kwargs = dict(system_prompt=system_prompt)
    if temperature is not None:
        # fallback to base class default if not passed
        kwargs["temperature"] = temperature

    if isinstance(input_audio, str):
        input_audio = Path(input_audio)

    if isinstance(input_audio, Path):
        return cls.from_file(
            model,
                input_audio,
                **kwargs
            )
    elif isinstance(input_audio, AudioSegment):
        return cls.from_audio(
            model,
            input_audio,
            **kwargs
        )
    else:
        raise ValueError(f"Invalid input type: {type(input_audio)}. Must be a file or AudioSegment.")

class SpokenWrapper:
    def __call__(self, model_name: str, input_f: Union[str, Path], system_prompt: Optional[str] = None, temperature: Optional[float] = None) -> SpeechToSpeechHarness:
        return spoken(model_name, input_f, system_prompt, temperature)

    @property
    def models(self) -> list[str]:
        return list(SpeechToSpeechHarnessMeta._model_registry.keys())

sys.modules[__name__] = SpokenWrapper()



import base64
import os
import sys
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Type

import numpy as np
from loguru import logger
from pydub import AudioSegment

if TYPE_CHECKING:
    from typing import Self

logger.remove()
logger.add(
    sys.stderr,
    format="<green>{time:HH:mm:ss}</green> | <level>{level: <8}</level> | <magenta>{extra[harness]}</magenta> - <level>{message}</level>",
    level=os.environ.get("LOG_LEVEL", "ERROR").upper(),
)

class SpeechToSpeechHarnessMeta(ABCMeta):
    _model_registry: Dict[str, Type['SpeechToSpeechHarness']] = {}

    def __new__(mcs, name, bases, namespace, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace)

        if 'Model' not in namespace or not issubclass(namespace['Model'], Enum):
            raise ValueError("Subclass of SpeechToSpeechHarness must expose available models in Model enum")

        for model in namespace['Model']:
            mcs._model_registry[model.value] = cls

        return cls

    def name_to_harness(model_name: str) -> Type['SpeechToSpeechHarness']:
        if model_name not in SpeechToSpeechHarness._model_registry:
            raise ValueError(f"Model {model_name} not found. Available models: {SpeechToSpeechHarness._model_registry.keys()}")
        return SpeechToSpeechHarness._model_registry[model_name]


class SpeechToSpeechHarness(ABC, metaclass=SpeechToSpeechHarnessMeta):
    class Model(Enum):
        pass

    # model-specific constants
    input_audio_sample_rate: Optional[int] = None # Hz
    output_audio_sample_rate: Optional[int] = None # Hz
    audio_token_frame_rate: Optional[int] = None  # Hz of the input audio -> # tokens

    source_audio_signal: np.ndarray

    input_audio_signal: np.ndarray
    input_audio_base64: str
    input_audio: Optional[AudioSegment] = None

    transcription: Optional[str] = None
    output_audio_bytes: bytes = b""
    output_audio: Optional[AudioSegment] = None

    model_name: str
    temperature: float

    # state
    is_ready: bool = False
    is_complete: bool = False

    input_audio_tokens: int
    input_transcription: Optional[str] = None

    output_audio_tokens: int
    output_transcription: Optional[str] = None

    # for latency analysis
    input_audio_sent_time: Optional[float] = None  # When we finish sending input audio
    first_output_token_time: Optional[float] = None  # When we receive first output token

    def __init__(
        self,
        model: Model,
        source_audio_signal: np.ndarray,
        system_prompt: Optional[str] = None,
        temperature: float = 0.8,
    ):
        self.model_name = model.value

        self.input_audio_signal = self.source_audio_signal = source_audio_signal
        self.input_audio_bytes = np.clip(self.input_audio_signal * (2**15), -32768, 32767).astype(np.int16).tobytes()
        self.input_audio_base64 = base64.b64encode(self.input_audio_bytes).decode("utf-8")

        self.system_prompt = system_prompt
        self.temperature = temperature

        self.logger = logger.bind(
            harness=self.__class__.__name__.split("SpeechToSpeech")[0]
        )

        self.input_audio_tokens = -1
        self.output_audio_tokens = -1

        self.is_ready = False
        self.is_complete = False

        self.output_audio_bytes = b""
        self.output_transcription = ""
        self.input_transcription = ""

        # latency analysis fields
        self.input_audio_sent_time = None
        self.first_output_token_time = None

    @classmethod
    def from_file(
        cls,
        model: Model,
        input_f: Path,
        system_prompt: Optional[str] = None,
        temperature: float = 0.8,
    ) -> 'SpeechToSpeechHarness':
        try:
            audio = AudioSegment.from_file(input_f)
            return cls.from_audio(model, audio, system_prompt, temperature)
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise Exception("Cannot parse source audio.", e)

    @classmethod
    def from_audio(
        cls,
        model: Model,
        input_audio: AudioSegment,
        system_prompt: Optional[str] = None,
        temperature: float = 0.8,
    ) -> 'SpeechToSpeechHarness':
        try:
            pcm_audio = (
                input_audio.set_frame_rate(cls.input_audio_sample_rate)
                .set_channels(1)
                .set_sample_width(2)
            )

            samples = np.array(pcm_audio.get_array_of_samples())
            source_audio_signal = samples.astype(np.float32) / (2**15)
        except Exception as e:
            import traceback
            traceback.print_exc()
            raise Exception("Cannot parse source audio.", e)

        return cls(model, source_audio_signal, system_prompt, temperature)

    @abstractmethod
    async def run(self) -> Tuple[Optional[str], str, AudioSegment]:
        """
        Kick off provider-specific control flow: setup, send input audio, receive output, and cleanup.

        Returns:
            input_transcription: Optional[str] - The transcription of the input audio.
            output_transcription: str - The transcription of the output audio.
            output_audio: AudioSegment - The output audio.
        """



import asyncio
import base64
import io
import json
import os
import time
from enum import Enum
from typing import Optional, Tuple

import numpy as np
import websockets
from pydub import AudioSegment

from spoken.base import SpeechToSpeechHarness


class GeminiSpeechToSpeechHarness(SpeechToSpeechHarness):
    """
    Gemini Multimodal Live (https://ai.google.dev/gemini-api/docs/live)

    Native audio (speech -> speech)
    - gemini-2.5-flash-preview-native-audio-dialog
    - gemini-2.5-flash-exp-native-audio-thinking-dialog
    """

    class Model(Enum):
        GEMINI_2_5_FLASH_PREVIEW_NATIVE_AUDIO_DIALOG = "gemini-2.5-flash-preview-native-audio-dialog"
        GEMINI_2_5_FLASH_EXP_NATIVE_AUDIO_THINKING_DIALOG = "gemini-2.5-flash-exp-native-audio-thinking-dialog"

    input_audio_sample_rate: Optional[int] = 16000
    output_audio_sample_rate: Optional[int] = 24000
    audio_token_frame_rate: Optional[int] = 25  # 40ms per token

    def __init__(
        self,
        model: Model,
        source_audio_signal: np.ndarray,
        system_prompt: Optional[str] = None,
        temperature: float = 1.0,
    ):
        super().__init__(model, source_audio_signal, system_prompt, temperature)

        api_key = os.environ.get("GEMINI_API_KEY")
        if not api_key:
            raise ValueError("GEMINI_API_KEY environment variable is required")

        self.realtime_base_url = f"wss://generativelanguage.googleapis.com/ws/google.ai.generativelanguage.v1beta.GenerativeService.BidiGenerateContent?key={api_key}"

    async def run(self) -> Tuple[Optional[str], str, AudioSegment]:
        async with websockets.connect(
            self.realtime_base_url,
        ) as websocket:
            setup_message = {
                "setup": {
                    "model": f"models/{self.model_name}",
                    "generationConfig": {
                        "temperature": self.temperature,
                        "responseModalities": ["AUDIO"]
                    },
                    "realtimeInputConfig": {
                        "automaticActivityDetection": {
                            "disabled": True
                        }
                    },
                    "inputAudioTranscription": {},
                    "outputAudioTranscription": {}
                }
            }

            if self.system_prompt:
                setup_message["setup"]["systemInstruction"] = self.system_prompt

            self.logger.debug(
                "Sending setup message: {}", json.dumps(setup_message, indent=2)
            )
            await websocket.send(json.dumps(setup_message))

            while not self.is_ready:
                try:
                    message = await asyncio.wait_for(websocket.recv(), timeout=1.0)
                    self._process_message(message)
                except asyncio.TimeoutError:
                    continue

            self.logger.debug("Sending audio input")
            await websocket.send(json.dumps({
                "realtimeInput": {
                    "activityStart": {}
                }
            }))

            audio_message = {
                "realtimeInput": {
                    "audio": {
                        "data": base64.urlsafe_b64encode(self.input_audio_bytes).decode('ascii'),
                        "mimeType": f"audio/pcm;rate={self.input_audio_sample_rate}"
                    },
                }
            }
            await websocket.send(json.dumps(audio_message))

            end_message = {
                "realtimeInput": {
                    "audioStreamEnd": True
                }
            }
            await websocket.send(json.dumps(end_message))

            self.input_audio_sent_time = time.perf_counter()
            self.logger.info("Input audio fully sent at: {}", self.input_audio_sent_time)

            while not self.is_complete:
                try:
                    message = await asyncio.wait_for(websocket.recv(), timeout=1.0)
                    self._process_message(message)
                except asyncio.TimeoutError:
                    continue

        self.output_audio = AudioSegment.from_raw(
            io.BytesIO(self.output_audio_bytes),
            sample_width=2,
            frame_rate=self.output_audio_sample_rate,
            channels=1,
        )

        return self.input_transcription, self.output_transcription, self.output_audio

    def _process_message(self, message: str):
        try:
            data = json.loads(message)
            self.logger.debug("Received event: {}", json.dumps(data, indent=2))

            if "setupComplete" in data:
                self.logger.info("Setup complete")
                self.is_ready = True

            elif "serverContent" in data:
                self.logger.info("Received server content")
                server_content = data["serverContent"]

                if "modelTurn" in server_content:
                    for part in server_content["modelTurn"].get("parts", []):
                        if "inlineData" in part and "audio/pcm" in (mime_type := part["inlineData"]["mimeType"]):
                            if self.first_output_token_time is None:
                                self.first_output_token_time = time.perf_counter()
                                self.logger.info("First output token received at: {}", self.first_output_token_time)

                            rate = int(mime_type.split(";rate=")[1]) if ";rate=" in mime_type else self.output_audio_sample_rate
                            self.output_audio_bytes += base64.b64decode(part["inlineData"]["data"])
                            self.logger.debug("Received audio chunk: {} bytes, rate: {}", len(part["inlineData"]["data"]), rate)

                if "inputTranscription" in server_content:
                    self.input_transcription += server_content["inputTranscription"]["text"]
                    self.logger.info("Input transcription: {}", self.input_transcription)

                if "outputTranscription" in server_content:
                    self.output_transcription += server_content["outputTranscription"]["text"]
                    self.logger.info("Transcription: {}", self.output_transcription)

                if server_content.get("turnComplete", False):
                    self.is_complete = True
                    self.logger.info("Turn complete")

                    usage_metadata = server_content.get("usageMetadata", {})
                    for token_details in usage_metadata.get("responseTokensDetails", []):
                        if token_details.get("modality") == "AUDIO":
                            self.output_audio_tokens = token_details.get("tokenCount", 0)
                            self.logger.info("Output tokens: {}", self.output_audio_tokens)

            elif "error" in data:
                self.logger.error("Server error: {}", json.dumps(data, indent=2))
                self.is_complete = True

        except json.JSONDecodeError as e:
            self.logger.error("Failed to parse message: {}", e)
        except Exception as e:
            self.logger.error("Error processing message: {}", e)

if __name__ == "__main__":
    from pathlib import Path

    # os.environ["GEMINI_API_KEY"] = "your-gemini-api-key"

    harness = GeminiSpeechToSpeechHarness.from_file(
        GeminiSpeechToSpeechHarness.Model.GEMINI_2_5_FLASH_PREVIEW_NATIVE_AUDIO_DIALOG,
        Path("./examples/scooby.wav")
    )

    input_transcription, output_transcription, output_audio = asyncio.run(harness.run())
    print(f"Input transcription: {input_transcription}")
    print(f"Output transcription: {output_transcription}")
    output_audio.export("./gemini.wav", format="wav")
